import numpy as np
import soundfile as sf
import matplotlib.pyplot as plt
from numpy.fft import fft
from numpy.fft import ifft

# Input
x = np.zeros((2048,))
x[0] = 1

# Read in prototype filter
g_alias, samplerate = sf.read('g.wav')
g_image = np.flip(g_alias)
N = len(g_alias)

# Parameters
M = 16
r = 12

# Allocation
# For understanding dimensions

# Input delay chain
input_buffer = np.zeros((N,))
# Result of alias filtering
weighted_alias = np.zeros((M,1))
# Final subband signals
subbands = np.zeros((M,),dtype=complex)
# Subbands after synthesis ifft
subbands_transformed = np.zeros((M,))
# Result of image filtering
weighted_image = np.zeros((N,1))
# Output delay chain
output_buffer = np.zeros((N,))
# Output signal
out = np.zeros((len(x),))

# Manages addition / extraction of weighted samples 
mapping = np.zeros((M,N))

# Set 1s in necessary positions
for i in range(M):
    for j in range(N):

        if (j-i) % M == 0:
            mapping[i,j] = 1

# Original sample rate
for i in range(len(x)):

    # Fill delay chain
    input_buffer[1:] = input_buffer[0:-1]
    input_buffer[0] = x[i]

    # Manage output delay chain
    # Upsampling is implicit
    output_buffer[0:-1] = output_buffer[1:]
    output_buffer[-1] = 0

    # Rate reduced by factor r
    if (i+1) % r == 0:

        # Weighting and mapping into M channels
        weighted_alias = mapping @ np.multiply(input_buffer,g_alias)

        # Apply analysis transform
        subbands = M*ifft(weighted_alias)

        # Room for subband processing

        # Apply synthesis transform, enforce real numbers
        subbands_transformed = np.real(ifft(subbands))

        # Mapping to N channels and weighting
        weighted_image = np.multiply(g_image,np.transpose(mapping) @ subbands_transformed)

        # Mix with output buffer
        output_buffer = output_buffer +  weighted_image

    # Extract output sample
    out[i] = output_buffer[0]

# Plot Spectrum of input and output
plt.plot(20*np.log10(np.abs(fft(x))))
plt.plot(20*np.log10(np.abs(fft(out))))

plt.title('Magnitude spectrum')
plt.xlabel('index')
plt.ylabel('magnitude spectrum [dB]')
plt.legend(['Kronecker','Output'])
plt.show(block='true')

# Plot unwrapped phase
plt.phase_spectrum(x,8000)
plt.phase_spectrum(g_alias,8000)
plt.phase_spectrum(g_image,8000)
plt.phase_spectrum(out,8000)

plt.title('Phase spectrum')
plt.legend(['in','anti-alias','ant-image','out'])
plt.show(block='true')
